from torch import nn, Tensor
from zeta.nn import MambaBlock, OutputHead
from zeta.nn.attention.multiquery_attention import MultiQueryAttention
import torch.nn.functional as F


class MambaFormerBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        d_state: int,
        d_conv: int,
        heads: int,
        dim_head: int,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.dim = dim
        self.d_state = d_state
        self.d_conv = d_conv
        self.heads = heads
        self.dim_head = dim_head

        # MambaBlock
        self.input_mamba = MambaBlock(
            dim, 1, d_state, d_conv, *args, **kwargs
        )

        # MultiQueryAttention
        self.attn = MultiQueryAttention(dim, heads, *args, **kwargs)

    def forward(self, x: Tensor):
        # First step is Attention
        skip = x

        # Attention
        attended, _, _ = self.attn(x)

        # Add residual
        x = attended + skip

        # Second step is MambaBlock
        skip_two = x

        # Mamba
        x = self.input_mamba(x)

        # Add residual and return
        return x + skip_two


class MambaFormer(nn.Module):
    def __init__(
        self,
        dim: int,
        num_tokens: int,
        depth: int,
        d_state: int,
        d_conv: int,
        heads: int,
        dim_head: int,
        return_tokens: bool = True,
        cross_entropy_ignore_index=0,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.dim = dim
        self.num_tokens = num_tokens
        self.depth = depth
        self.d_state = d_state
        self.d_conv = d_conv
        self.heads = heads
        self.dim_head = dim_head
        self.return_tokens = return_tokens
        self.cross_entropy_ignore_index = cross_entropy_ignore_index

        # Embedding
        self.embed = nn.Embedding(num_tokens, dim)

        # Mamba as input
        self.input_mamba = MambaBlock(
            dim, 1, d_state, d_conv, *args, **kwargs
        )

        # Layers of Mamba Former Block
        self.layers = nn.ModuleList(
            [
                MambaFormerBlock(
                    dim,
                    d_state,
                    d_conv,
                    heads,
                    dim_head,
                )
                for _ in range(depth)
            ]
        )

        # Norm
        self.norm = nn.LayerNorm(dim)

    def forward(
        self,
        x,
        return_loss: bool = False,
        return_embeddings: bool = False,
    ):
        if return_loss:
            x, labels = x[:, -1], x[:, 1:]

        mask = x >= 0
        x = x.masked_fill(~mask, 0)

        # Embed tokens to get tensors
        x = self.embed(x)

        # Normalize
        x = self.norm(x)

        # Then skip
        skip = x

        # Mamba
        x = self.input_mamba(x) + skip

        # MambaFormer Blocks
        for layer in self.layers:
            x = layer(x) + x

        if self.return_tokens:
            out = OutputHead(self.dim, -1)(x)
            return out
        elif return_loss:
            logits = OutputHead(self.dim, -1)(x)
            out = F.cross_entropy(
                logits,
                labels,
                ignore_index=self.cross_entropy_ignore_index,
            )
        elif return_embeddings:
            return x



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mamba_former']

package_data = \
{'': ['*']}

install_requires = \
['einops', 'torch', 'zetascale==2.2.7']

setup_kwargs = {
    'name': 'mamba-former',
    'version': '0.0.3',
    'description': 'Paper - Pytorch',
    'long_description': '[![Multi-Modality](agorabanner.png)](https://discord.gg/qUtxnK2NMf)\n\n# MambaFormer\nImplementation of MambaFormer in Pytorch ++ Zeta from the paper: "Can Mamba Learn How to Learn? A Comparative Study on In-Context Learning Tasks"\n\n## install\n`pip3 install mamba-former`\n\n## usage\n```python\nimport torch\nfrom mamba_former.main import MambaFormer\n\n# Forward pass example\nx = torch.randint(1, 1000, (1, 100))  # Token\n# Tokens are integers representing input data\n\n# Model\nmodel = MambaFormer(\n    dim=512,  # Dimension of the model\n    num_tokens=1000,  # Number of unique tokens in the input data\n    depth=6,  # Number of transformer layers\n    d_state=512,  # Dimension of the transformer state\n    d_conv=128,  # Dimension of the convolutional layer\n    heads=8,  # Number of attention heads\n    dim_head=64,  # Dimension of each attention head\n    return_tokens=True,  # Whether to return the tokens in the output\n)\n\n# Forward pass\nout = model(x)  # Perform a forward pass through the model\n\n# If training\n# out = model(x, return_loss=True)  # Perform a forward pass and calculate the loss\n\n# Print the output\nprint(out)  # Print the output tensor\nprint(out.shape)  # Print the shape of the output tensor\n\n```\n\n\n# License\nMIT\n',
    'author': 'Kye Gomez',
    'author_email': 'kye@apac.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kyegomez/MambaFormer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
